/*
 * KernelConf - A Linux Kernel Configuration System
 *
 * Distributed under GNU General Public Lisence
 *
 * Copyright (C) 2002 by Anuradha Ratnaweera
 */

#include <ctype.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "kernelconf.h"
#include "expressions.h"
#include "utils.h"
#include "tree.h"

extern IdPtr id_tree_root;

/*
 * Read the next token from the expression and returns its type
 * precedence and length.  Apart from the operator types, SYMBOL
 * and VALUE types are recognized.  Symbols contain alphabatical,
 * numerical and underscore characters and values are contained
 * within quotes.  TYPE_NULL is returned if we are at the end of
 * the string.
 */
static char *get_next_token(char* expr, TokenPtr token, size_t *length)
{
    char *p;

    while (isspace(*expr))
	expr++;

    switch (*expr) {
	case 0:
	    *length = 1;
	    token->type = TYPE_NULL;
	    token->precedence = PRE_NULL;
	    return expr;
	case '<':
	    switch (*(expr + 1)) {
		case '-':
		    *length = 2;
		    token->type = TYPE_DERIVED;
		    token->precedence = PRE_ASSIGN;
		    return expr;
		case '=':
		    *length = 2;
		    token->type = TYPE_LTE;
		    token->precedence = PRE_REL;
		    return expr;
		default:
		    *length = 1;
		    token->type = TYPE_LT;
		    token->precedence = PRE_REL;
		    return expr;
	    }
	case '>':
	    switch (*(expr + 1)) {
		case '=':
		    *length = 2;
		    token->type = TYPE_GTE;
		    token->precedence = PRE_REL;
		    return expr;
		default:
		    *length = 1;
		    token->type = TYPE_GT;
		    token->precedence = PRE_REL;
		    return expr;
	    }
	case '(':
	    *length = 1;
	    token->type = TYPE_LEFT_PARA;
	    token->precedence = PRE_NULL;
	    return expr;
	case ')':
	    *length = 1;
	    token->type = TYPE_RIGHT_PARA;
	    token->precedence = PRE_NULL;
	    return expr;
	case '!':
	    switch (*(expr + 1)) {
		case '=':
		    *length = 2;
		    token->type = TYPE_NOT_EQUAL;
		    token->precedence = PRE_EQUAL;
		    return expr;
		default:
		    *length = 1;
		    token->type = TYPE_NOT;
		    token->precedence = PRE_NOT;
		    return expr;
	    }
	case '=':
	    switch (*(expr + 1)) {
		case '=':
		    *length = 2;
		    token->type = TYPE_EQUAL;
		    token->precedence = PRE_EQUAL;
		    return expr;
		default:
		    *length = 0;
		    token->type = TYPE_UNKNOWN;
		    token->precedence = PRE_NULL;
		    return expr;
	    }
	case '&':
	    switch (*(expr + 1)) {
		case '&':
		    *length = 2;
		    token->type = TYPE_AND;
		    token->precedence = PRE_AND;
		    return expr;
		default:
		    *length = 0;
		    token->type = TYPE_UNKNOWN;
		    token->precedence = PRE_NULL;
		    return expr;
	    }
	case '|':
	    switch (*(expr + 1)) {
		case '|':
		    *length = 2;
		    token->type = TYPE_OR;
		    token->precedence = PRE_OR;
		    return expr;
		default:
		    *length = 0;
		    token->type = TYPE_UNKNOWN;
		    token->precedence = PRE_NULL;
		    return expr;
	    }

	case '-':
	    switch (*(expr + 1)) {
		case '>':
		    *length = 2;
		    token->type = TYPE_DEPENDS;
		    token->precedence = PRE_ASSIGN;
		    return expr;
		default:
		    *length = 0;
		    token->type = TYPE_UNKNOWN;
		    token->precedence = PRE_NULL;
		    return expr;
	    }
	case '"':
	    p = ++expr;
	    while ((*p) && ((*p) != '"'))
		p++;
	    if (*p) {
		*p = ' ';
		*length = p - expr;
		token->type = TYPE_VALUE;
		token->precedence = PRE_NULL;
		return expr;
	    }
	    else {
		*length = 0;
		token->type = TYPE_UNKNOWN;
		token->precedence = PRE_NULL;
		return expr;
	    }
	default:
	    p = expr;
	    while (isalnum(*p) || ((*p) == '_'))
		p++;
	    if (p == expr) {
		*length = 0;
		token->type = TYPE_UNKNOWN;
		token->precedence = PRE_NULL;
		return expr;
	    }
	    else {
		*length = p - expr;
		token->type = TYPE_SYMBOL;
		token->precedence = PRE_NULL;
		return expr;
	    }
    }

    *length = 0;
    token->type = TYPE_UNKNOWN;
    token->precedence = PRE_NULL;
    return expr;
}

/*
 * Pushes the TOKEN into the stack.  *STACK == NULL if empty.
 */
static void push(TokenPtr *stack, TokenPtr token)
{
    token->next = *stack;
    *stack = token;
}

/*
 * Pops from the stack and returns a pointer to the top element.
 * Returns NULL if the stack is empty.
 */
static TokenPtr pop(TokenPtr *stack)
{
    TokenPtr token;

    if (*stack) {
	token = *stack;
	*stack = (*stack)->next;
	token->next = NULL;
	return token;
    }
    else
	return NULL;
}

#ifdef DEBUGGED

static void show_list(TokenPtr list, char *str)
{
    TokenPtr token;
    printf("%s: ", str);
    for (token = list; token; token = token->next)
	if (token->type == TYPE_VALUE)
	    printf("\"%s\" ", (char*)(token->value.str_val));
	else if (token->type == TYPE_SYMBOL)
	    printf("*%s* ", token->value.id_val->name);
	else if (token->type == TYPE_BOOL)
	    printf("*%c* ", token->value.char_val);
	else if (token->type == TYPE_TRI)
	    printf("*%c* ", token->value.char_val);
	else
	    printf("%d ", token->type);
    printf("\n");
}

#endif

/*
 * Adds TOKEN to the end of a linked list pointed by FIRST whose last
 * element is LAST.
 */
static void add_token_to_list(TokenPtr *first, TokenPtr *last, TokenPtr token)
{
    if (*first)
	(*last)->next = token;
    else
	*first = token;

    *last = token;
	token->next = NULL;
}

/*
 * If the token has any memory allocated, free it
 */
static void clear_token(TokenPtr token)
{
    if ((token->type == TYPE_VALUE) || (token->type == TYPE_STR))
	free(token->value.str_val);
}

/*
 * Releases memory used by token and if necessary by its elements.
 */
void free_token(TokenPtr token)
{
    clear_token(token);
    free(token);
}

/*
 * Parses the string EXPR related to SYMBOL and returns an expression
 * which contains a list of tokens in reverse Polish.
 */
static ExpressionPtr parse_expression(char *expr)
{
    ExpressionPtr expression;
    TokenPtr stack, token, last_token, tmp;
    size_t length;
    char c;

    expression = malloc(sizeof(Expression));
    expression->first_token = NULL;
    expression->next = NULL;

    stack = NULL;

    do {
	token = (TokenPtr)malloc(sizeof(Token));
	expr = get_next_token(expr, token, &length);

	switch (token->type) {
	    case TYPE_LEFT_PARA:
		push(&stack, token);
		break;
	    case TYPE_RIGHT_PARA:
		while (stack && (stack->type != TYPE_LEFT_PARA)) {
		    tmp = pop(&stack);
		    add_token_to_list(&(expression->first_token),
			    &last_token, tmp);
#ifdef DEBUGGED
		    show_list(stack, "Stack");
		    show_list(expression->first_token, "Expression");
#endif
		}
		pop(&stack);
		break;
	    case TYPE_OR:
	    case TYPE_AND:
	    case TYPE_EQUAL:
	    case TYPE_NOT_EQUAL:
	    case TYPE_LT:
	    case TYPE_GT:
	    case TYPE_LTE:
	    case TYPE_GTE:
	    case TYPE_NOT:
	    case TYPE_DEPENDS:
	    case TYPE_DERIVED:
		while (stack && (stack->precedence >= token->precedence)) {
		    tmp = pop(&stack);
		    add_token_to_list(&(expression->first_token),
			    &last_token, tmp);
#ifdef DEBUGGED
		    show_list(stack, "Stack");
		    show_list(expression->first_token, "Expression");
#endif
		}
		push(&stack, token);
		break;
	    case TYPE_SYMBOL:
		c = expr[length];
		expr[length] = 0;
		token->value.id_val = search_tree(&id_tree_root, expr,
			ID_MARK_INSERT);
		add_token_to_list(&(expression->first_token),
			&last_token, token);
		expr[length] = c;
		break;
	    case TYPE_VALUE:
		c = expr[length];
		expr[length] = 0;
		token->value.str_val = string_duplicate(expr);
		add_token_to_list(&(expression->first_token),
			&last_token, token);
		expr[length] = c;
		break;
	    case TYPE_UNKNOWN:
		return NULL;
	    case TYPE_NULL:
		/* why doesn't the following line work?
		free(token); */
		break;
	    default:
		;
	}

#ifdef DEBUGGED
	show_list(stack, "Stack");
	show_list(expression->first_token, "Expression");
#endif

	expr += length;
    } while (token->type != TYPE_NULL);

    while (stack) {
	tmp = pop(&stack);
	add_token_to_list(&(expression->first_token),
		&last_token, tmp);
#ifdef DEBUGGED
	show_list(stack, "Stack");
	show_list(expression->first_token, "Expression");
#endif
    }

    return expression;
}

/*
 * Accepts a semicolon seperated list of expressions EXPR, parses each one by
 * repeatedly calling parse_expression, and adds to the end of CURRENT_LIST.
 */
ExpressionPtr parse_expressions(char *expr, ExpressionPtr current_list)
{
    char *c;
    ExpressionPtr expression_list, last_expression, expression;

    if (current_list) {
	expression_list = current_list;
	last_expression = current_list;
	while (last_expression->next)
	    last_expression = last_expression->next;
    }
    else {
	expression_list = NULL;
    }

    do {
	c = strchr(expr, ';');
	if (c)
	    *c = 0;
#ifdef DEBUGGED
	printf("Evaluating: *%s*\n", expr);
#endif
	expression = parse_expression(expr);
	if (expression) {
	    expression->string = string_duplicate(expr);
	    if (expression_list) {
		last_expression->next = expression;
		last_expression = expression;
	    }
	    else {
		expression_list = expression;
		last_expression = expression;
	    }
	}

	if (c)
	    expr = c + 1;
    } while (c);
    
    return expression_list;
}

/*
 * End of parsing expressions.
 * Now we've got to evaluate them.
 */

/*
 * Get the boolean compatible value of a token.
 */
char get_boolean_value(TokenPtr token)
{
    switch (token->type) {
	case TYPE_BOOL:
	case TYPE_TRI:
	case TYPE_OPTION:
	    return token->value.bool_val;
	case TYPE_VALUE:
	    if (((token->value.str_val[0] == 'n') ||
			(token->value.str_val[0] == 'm') ||
			(token->value.str_val[0] == 'y')) &&
		    (token->value.str_val[1] == 0))
		return token->value.str_val[0];
	    else
		return 'n';
	case TYPE_INT:
	case TYPE_HEX:
	    if (token->value.int_val)
		return 'y';
	    else
		return 'n';
	case TYPE_STR:
	    if (token->value.str_val[0])
		return 'y';
	    else
		return 'n';
	case TYPE_NULL: /* just to show that NULL is boolean "n" */
	default:
	    return 'n';
    }
}

static void do_operation_or(TokenPtr *stack)
{
    TokenPtr t1, t2;
    char result, v1, v2;

    t2 = pop(stack);
    t1 = pop(stack);
    v1 = get_boolean_value(t1);
    v2 = get_boolean_value(t2);
    clear_token(t1);
    free_token(t2);

    if ((v1 == 'y') || (v2 == 'y'))
	result = 'y';
    else if ((v1 == 'm') || (v2 == 'm'))
	result = 'm';
    else
	result = 'n';

    t1->type = TYPE_TRI;
    t1->value.bool_val = result;
    t1->precedence = PRE_NULL;
    push(stack, t1);
}

static void do_operation_and(TokenPtr *stack)
{
    TokenPtr t1, t2;
    char result, v1, v2;

    t2 = pop(stack);
    t1 = pop(stack);
    v1 = get_boolean_value(t1);
    v2 = get_boolean_value(t2);
    clear_token(t1);
    free_token(t2);

    if ((v1 == 'n') || (v2 == 'n'))
	result = 'n';
    else if ((v1 == 'm') || (v2 == 'm'))
	result = 'm';
    else
	result = 'y';

    t1->type = TYPE_TRI;
    t1->value.bool_val = result;
    t1->precedence = PRE_NULL;
    push(stack, t1);
}

static void do_operation_not(TokenPtr *stack)
{
    TokenPtr t;
    char result, v;

    t = pop(stack);
    v = get_boolean_value(t);
    clear_token(t);

    if ((v == 'y') || (v == 'm'))
	result = 'n';
    else
	result = 'y';

    t->type = TYPE_BOOL;
    t->value.bool_val = result;
    t->precedence = PRE_NULL;
    push(stack, t);
}

Type get_generic_type(Type type)
{
    switch (type) {
	case TYPE_BOOL:
	case TYPE_TRI:
	case TYPE_OPTION:
	    return TYPE_TRI;
	case TYPE_INT:
	    return TYPE_INT;
	case TYPE_HEX:
	    return TYPE_HEX;
	case TYPE_CHAR:
	    return TYPE_CHAR;
	case TYPE_STR:
	    return TYPE_STR;
	case TYPE_VALUE:
	    return TYPE_VALUE;
	case TYPE_NULL:
	    return TYPE_NULL;
	default:
	    return TYPE_UNKNOWN;
    }
}

static void set_types_to_compare(TokenPtr t1, TokenPtr t2)
{
    int i;
    char c;

    t1->type = get_generic_type(t1->type);
    t2->type = get_generic_type(t2->type);

    if ((t1->type == TYPE_VALUE) && (t2->type != TYPE_VALUE)) {
	if (t2->type == TYPE_TRI) {
	    c = get_boolean_value(t1);
	    free(t1->value.str_val);
	    t1->value.bool_val = c;
	    t1->type = TYPE_TRI;
	}
	else if (t2->type == TYPE_INT) {
	    i = atoi(t1->value.str_val);
	    free(t1->value.str_val);
	    t1->value.int_val = i;
	    t1->type = TYPE_INT;
	}
	else if (t2->type == TYPE_HEX) {
	    sscanf(t1->value.str_val, "%x", &i);
	    free(t1->value.str_val);
	    t1->value.int_val = i;
	    t1->type = TYPE_INT;
	}
    }
    if ((t1->type != TYPE_VALUE) && (t2->type == TYPE_VALUE)) {
	if (t1->type == TYPE_TRI) {
	    c = get_boolean_value(t2);
	    free(t2->value.str_val);
	    t2->value.bool_val = c;
	    t2->type = TYPE_TRI;
	}
	else if (t1->type == TYPE_INT) {
	    /* handle hex values seperately */
	    i = atoi(t2->value.str_val);
	    free(t2->value.str_val);
	    t2->value.int_val = i;
	    t2->type = TYPE_INT;
	}
	else if (t1->type == TYPE_HEX) {
	    sscanf(t2->value.str_val, "%x", &i);
	    free(t2->value.str_val);
	    t2->value.int_val = i;
	    t2->type = TYPE_INT;
	}
    }

    /* FIXME: handle other types */
}

static void do_operation_equal(TokenPtr *stack)
{
    TokenPtr t1, t2;
    char result;

    t2 = pop(stack);
    t1 = pop(stack);

    set_types_to_compare(t1, t2);

    result = 'n';
    if (t1->type == t2->type) {
	if (t1->type == TYPE_TRI) {
	    if (t1->value.bool_val == t2->value.bool_val)
		result = 'y';
	}
	else if (t1->type == TYPE_INT) {
	    if (t1->value.int_val == t2->value.int_val)
		result = 'y';
	}
	else if (t1->type == TYPE_CHAR) {
	    if (t1->value.char_val == t2->value.char_val)
		result = 'y';
	}
	else if (t1->type == TYPE_STR) {
	    if (strcmp(t1->value.str_val, t2->value.str_val) == 0)
		result = 'y';
	}
	else if (t1->type == TYPE_NULL) {
	    result = 'y';
	}
    }

    clear_token(t1);
    t1->type = TYPE_BOOL;
    t1->value.bool_val = result;
    t1->precedence = PRE_NULL;
    push(stack, t1);

    free_token(t2);
}

static void do_operation_not_equal(TokenPtr *stack)
{
    do_operation_equal(stack);
    do_operation_not(stack);
}

static void do_operation_lt(TokenPtr *stack)
{
    TokenPtr t1, t2;
    char result;

    t2 = pop(stack);
    t1 = pop(stack);

    set_types_to_compare(t1, t2);

    result = 0;
    if (t1->type == t2->type) {
	if (t1->type == TYPE_TRI) {
	    if (((t1->value.bool_val == 'n') && ((t2->value.bool_val == 'm') ||
			    (t2->value.bool_val == 'y'))) ||
		    ((t1->value.bool_val == 'm') &&
		     (t2->value.bool_val == 'y')))
		result = 'y';
	    else
		result = 'n';
	}
	else if (t1->type == TYPE_INT) {
	    if (t1->value.int_val < t2->value.int_val)
		result = 'y';
	    else
		result = 'n';
	}
	else if (t1->type == TYPE_CHAR) {
	    if (t1->value.char_val < t2->value.char_val)
		result = 'y';
	    else
		result = 'n';
	}
	else if (t1->type == TYPE_STR) {
	    if (strcmp(t1->value.str_val, t2->value.str_val) < 0)
		result = 'y';
	    else
		result = 'n';
	}
    }

    clear_token(t1);
    if (result) {
	t1->type = TYPE_BOOL;
	t1->value.bool_val = result;
    }
    else
	t1->type = TYPE_NULL;
    t1->precedence = PRE_NULL;
    push(stack, t1);
    free_token(t2);
}

static void do_operation_gt(TokenPtr *stack)
{
    TokenPtr t1, t2;
    char result;

    t2 = pop(stack);
    t1 = pop(stack);

    set_types_to_compare(t1, t2);

    result = 0;
    if (t1->type == t2->type) {
	if (t1->type == TYPE_TRI) {
	    if (((t1->value.bool_val == 'y') && ((t2->value.bool_val == 'n') ||
			    (t2->value.bool_val == 'm'))) ||
		    ((t1->value.bool_val == 'y') &&
		     (t2->value.bool_val == 'm')))
		result = 'y';
	    else
		result = 'n';
	}
	else if (t1->type == TYPE_INT) {
	    if (t1->value.int_val > t2->value.int_val)
		result = 'y';
	    else
		result = 'n';
	}
	else if (t1->type == TYPE_CHAR) {
	    if (t1->value.char_val > t2->value.char_val)
		result = 'y';
	    else
		result = 'n';
	}
	else if (t1->type == TYPE_STR) {
	    if (strcmp(t1->value.str_val, t2->value.str_val) > 0)
		result = 'y';
	    else
		result = 'n';
	}
    }

    if (t1->type == TYPE_VALUE)
	free(t1->value.str_val);
    if (result) {
	t1->type = TYPE_BOOL;
	t1->value.bool_val = result;
    }
    else
	t1->type = TYPE_NULL;
    t1->precedence = PRE_NULL;
    push(stack, t1);
    free_token(t2);
}

static void do_operation_lte(TokenPtr *stack)
{
    do_operation_gt(stack);
    if ((*stack)->type != TYPE_NULL)
	do_operation_not(stack);
}

static void do_operation_gte(TokenPtr *stack)
{
    do_operation_lt(stack);
    if ((*stack)->type != TYPE_NULL)
	do_operation_not(stack);
}

static void do_operation_depends(TokenPtr *stack)
{
    TokenPtr t1, t2;
    char result, v1, v2;

    t2 = pop(stack);
    t1 = pop(stack);
    v1 = get_boolean_value(t1);
    v2 = get_boolean_value(t2);
    clear_token(t1);

    if ((v1 == 'y') && (v2 == 'n'))
	result = 'n';
    else
	result = 'y';

    t1->type = TYPE_BOOL;
    t1->value.bool_val = result;
    t1->precedence = PRE_NULL;
    push(stack, t1);
    free_token(t2);
}

static void do_operation_derived(TokenPtr *stack)
{
    TokenPtr t;
    char v;

    t = pop(stack);
    v = get_boolean_value(t);
    free_token(t);

    if (v == 'y')
	return;

    t = pop(stack);
    t->type = TYPE_NULL;
    t->precedence = PRE_NULL;
    push(stack, t);
}

static TokenPtr duplicate_token(TokenPtr token)
{
    TokenPtr tmp;

    tmp = (TokenPtr)malloc(sizeof(Token));
    tmp->type = token->type;
    tmp->precedence = token->precedence;
    if ((tmp->type == TYPE_VALUE) || (tmp->type == TYPE_STR))
	tmp->value.str_val = string_duplicate(token->value.str_val);
    else
	tmp->value = token->value;
    tmp->next = token->next;

    return tmp;
}

/*
 * Expressions are represented as a sequence of tokens.  These tokens are
 * operators, of TYPE_VALUE or of type TYPE_SYMBOL.  While evaluating, we don't
 * alter these lists, but work on copies of tokens.  However, while copying
 * tokens of TYPE_SYMBOL we also copy their type and value.
 */
TokenPtr evaluate_expression(ExpressionPtr expression)
{
    TokenPtr token, stack, t;

    stack = NULL;
    
    for (t = expression->first_token; t; t = t->next) {
	token = duplicate_token(t);
	switch (t->type) {
	    case TYPE_OR:
		do_operation_or(&stack);
		free_token(token);
		break;
	    case TYPE_AND:
		do_operation_and(&stack);
		free_token(token);
		break;
	    case TYPE_EQUAL:
		do_operation_equal(&stack);
		free_token(token);
		break;
	    case TYPE_NOT_EQUAL:
		do_operation_not_equal(&stack);
		free_token(token);
		break;
	    case TYPE_LT:
		do_operation_lt(&stack);
		free_token(token);
		break;
	    case TYPE_GT:
		do_operation_gt(&stack);
		free_token(token);
		break;
	    case TYPE_LTE:
		do_operation_lte(&stack);
		free_token(token);
		break;
	    case TYPE_GTE:
		do_operation_gte(&stack);
		free_token(token);
		break;
	    case TYPE_NOT:
		do_operation_not(&stack);
		free_token(token);
		break;
	    case TYPE_DEPENDS:
		do_operation_depends(&stack);
		free_token(token);
		break;
	    case TYPE_DERIVED:
		do_operation_derived(&stack);
		free_token(token);
		break;
	    case TYPE_SYMBOL:
		token->type = t->value.id_val->symbol->type;
		token->value = t->value.id_val->symbol->value;
		/* do not break */
	    case TYPE_VALUE:
		push(&stack, token);
		break;
	    default:
		;
	}
    }
    token = pop(&stack);
    return token;
}

