/*
 * KernelConf - A Linux Kernel Configuration System
 *
 * Distributed under GNU General Public Lisence
 *
 * Copyright (C) 2002 by Anuradha Ratnaweera
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "../include/linux/version.h"

#include "tree.h"
#include "deps.h"
#include "expressions.h"
#include "help.h"
#include "writeconfig.h"

extern IdPtr root_menu, id_tree_root;

#define MAX_MENU_NAME_LENGTH 256

#define BACK_TITLE "Linux Kernel V" UTS_RELEASE " Configuration"

#define MAX_INDENT 30   /* should be even */

static char indent_string[MAX_INDENT+1];
static char child_string[MAX_INDENT+1];
static int rows, cols;

static int set_geometry()
{
    FILE *stty_out;
    system("stty size > stty-out 2> /dev/null");
    if ((stty_out = fopen("stty-out", "rt")) == NULL)
	return 1;
    fscanf(stty_out, "%d %d", &rows, &cols);
    fclose(stty_out);
    unlink("stty-out");
    return 0;
}

/* +1 increases, -1 decreases, 0 resets, +2 increases with bar | */
static void set_indent_string(int indent)
{
    static int indent_ptr = 0;

    if ((indent_ptr == MAX_INDENT) && (indent > 0))
	return;

    switch (indent) {
	case 1:
	    indent_string[indent_ptr++] = ' ';
	    indent_string[indent_ptr++] = ' ';
	    break;
	case 2:
	    indent_string[indent_ptr++] = ' ';
	    indent_string[indent_ptr++] = '|';
	    break;
	case -1:
	    if (indent_ptr > 0)
		indent_ptr -= 2;
	    break;
	default:
	    indent_ptr = 0;
    }

    indent_string[indent_ptr] = 0;
}

/* 1 "+-", 2 "`=", 0 resets */
static void set_child_string(int type)
{
    switch (type) {
	case 1:
	    child_string[0] = ' ';
	    child_string[1] = '+';
	    child_string[2] = 0;
	    break;
	case 2:
	    child_string[0] = ' ';
	    child_string[1] = '`';
	    child_string[2] = 0;
	    break;
	default:
	    child_string[0] = 0;
    }
}

static int show_help(IdPtr id)
{
    FILE *help_file, *mc_help;
    char *buffer;
    long bufsize;

    if ((bufsize = read_help(id, &buffer)) < 0)
	return 1;

    if ((help_file = fopen("help.out", "wt")) == NULL)
	return 2;
    fwrite(buffer, sizeof(char), bufsize, help_file);
    free(buffer);
    fclose(help_file);

    if ((mc_help = fopen("MChelp", "wt")) == NULL)
	return 3;
    fprintf(mc_help, "./conf/lxdialog/lxdialog --title '%s' --backtitle '" BACK_TITLE "' --textbox help.out %d %d",
	    id->menu->caption, rows - 1, cols - 1);
    fclose(mc_help);

    system("sh MChelp");

    unlink("help.out");
    unlink("MChelp");

    return 0;
}

static int get_input(IdPtr id)
{
    FILE *mc_in, *mc_out;
    int retval;

    do {
	if ((mc_in = fopen("MCin", "wt")) == NULL)
	    return 3;
	switch (id->symbol->type) {
	    case TYPE_INT:
		fprintf(mc_in, "./conf/lxdialog/lxdialog --title '%s' --backtitle '" BACK_TITLE "' --inputbox 'Please enter an integer value.  Use the <TAB> key to move from the input field to the buttons below it' %d %d %d",
			id->menu->caption, 10, cols - 1, id->symbol->value.int_val);
		break;
	    case TYPE_HEX:
		fprintf(mc_in, "./conf/lxdialog/lxdialog --title '%s' --backtitle '" BACK_TITLE "' --inputbox 'Please enter a hexadecimal value.  Use the <TAB> key to move from the input field to the buttons below it' %d %d %x",
			id->menu->caption, 10, cols - 1, id->symbol->value.int_val);
		break;
	    case TYPE_STR:
		fprintf(mc_in, "./conf/lxdialog/lxdialog --title '%s' --backtitle '" BACK_TITLE "' --inputbox 'Please enter a string value.  Use the <TAB> key to move from the input field to the buttons below it' %d %d '%s'",
			id->menu->caption, 10, cols - 1, id->symbol->value.str_val);
		break;

	    default:
		;
	}
	fclose(mc_in);
	retval = system("sh MCin 2> MCout");

	if (retval) show_help(id);
    } while (retval);

    if ((mc_out = fopen("MCout", "rt")) == NULL)
	return 4;

    switch (id->symbol->type) {
	case TYPE_INT:
	    fscanf(mc_out, "%d", &(id->symbol->value.int_val));
	    break;
	case TYPE_HEX:
	    fscanf(mc_out, "%x", &(id->symbol->value.int_val));
	    break;
	case TYPE_CHAR:
	    fscanf(mc_out, "%c", &(id->symbol->value.char_val));
	    break;
	default:
	    ;
    }

    fclose(mc_out);

    unlink("MCin");
    unlink("MCout");

    return 0;
}

static int print_menu_item(IdPtr id, FILE *mc_menu, int expanded)
{
    IdPtr option;
    char value;

    if (id->symbol) {
	switch (id->symbol->type) {
	    case TYPE_BOOL:
		if (id->symbol->value.char_val == 'y')
		    value = '*';
		else
		    value = ' ';
		fprintf(mc_menu, " '%s' '%s%s[%c] %s'", id->name, indent_string,
			child_string, value, id->menu->caption);
		break;
	    case TYPE_TRI:
		if (id->symbol->value.char_val == 'y')
		    value = '*';
		else if (id->symbol->value.char_val == 'm')
		    value = 'M';
		else
		    value = ' ';
		fprintf(mc_menu, " '%s' '%s%s<%c> %s'", id->name, indent_string,
			child_string, value, id->menu->caption);
		break;
	    case TYPE_INT:
		fprintf(mc_menu, " '%s' '%s%s(%d) %s'", id->name, indent_string,
			child_string, id->symbol->value.int_val, id->menu->caption);
		break;
	    case TYPE_HEX:
		fprintf(mc_menu, " '%s' '%s%s(%x) %s'", id->name, indent_string,
			child_string, id->symbol->value.int_val, id->menu->caption);
		break;
	    case TYPE_CHAR:
		fprintf(mc_menu, " '%s' '%s%s(%c) %s'", id->name, indent_string,
			child_string, id->symbol->value.char_val, id->menu->caption);
		break;
	    case TYPE_STR:
		fprintf(mc_menu, " '%s' '%s%s(%s) %s'", id->name, indent_string,
			child_string, id->symbol->value.str_val, id->menu->caption);
		break;
	    case TYPE_OPTION:
		if (id->symbol->value.char_val == 'y')
		    value = 'X';
		else
		    value = ' ';
		fprintf(mc_menu, " '%s' '%s%s(%c) %s'", id->name, indent_string,
			child_string, value, id->menu->caption);
		break;
	    default:
		break;
	}
    }
    else if (id->menu->first_child && id->menu->first_child->symbol &&
	    id->menu->first_child->symbol->type == TYPE_OPTION) {
	fprintf(mc_menu, " '%s' '%s%s(", id->name, indent_string, child_string);
	for (option = id->menu->first_child; option;
		option = option->menu->next)
	    if (option->symbol->value.bool_val == 'y') {
		fprintf(mc_menu, "%s", option->menu->caption);
		break;
	    }
	fprintf(mc_menu, ") %s'", id->menu->caption);
    }
    else {
	if (expanded)
	    fprintf(mc_menu, " '%s' '%s%s[^] %s'", id->name, indent_string,
		    child_string, id->menu->caption);
	else
	    fprintf(mc_menu, " '%s' '%s%s[v] %s'", id->name, indent_string,
		    child_string, id->menu->caption);
    }
    return 0;
}

static int print_menu(IdPtr parent, FILE *mc_menu, int depth)
{
    IdPtr item;
    int n, menu_ok;

    item = parent;
    n = depth;

    set_indent_string(0);
    set_child_string(0);

    while (item) {
	menu_ok = 1;
	menu_ok = evaluate_all_expressions(item->menu->depends);
	if (menu_ok && n && item->menu && item->menu->first_child &&
		(item->menu->flags & MENU_FLAG_EXPANDED)) {
	    print_menu_item(item, mc_menu, 1);
	    if (child_string[0]) {
		if (item->menu->next)
		    set_indent_string(2);
		else
		    set_indent_string(1);
	    }
	    item = item->menu->first_child;
	    n--;
	    if (item->menu->next)
		set_child_string(1);
	    else
		set_child_string(2);
	}
	else if (item->menu->next) {
	    if (menu_ok)
		print_menu_item(item, mc_menu, 0);
	    item = item->menu->next;
	    if (item->menu->next == NULL)
		set_child_string(2);
	}
	else if (item->menu->parent) {
	    if (menu_ok)
		print_menu_item(item, mc_menu, 0);
	    item = item->menu->parent->menu->next;
	    n++;
	    if (item && item->menu && item->menu->next)
		set_child_string(1);
	    else
		set_child_string(2);
	    set_indent_string(-1);
	}
	else {
	    if (menu_ok)
		print_menu_item(item, mc_menu, 0);
	    item = NULL;
	}
    }
    return 0;
}

static int do_menu(IdPtr parent)
{
    FILE *mc_menu, *mc_out;
    IdPtr current, option, current0;
    IdListPtr id;
    char menu_name[MAX_MENU_NAME_LENGTH];
    int retval;

    if (parent->symbol)
	current = parent;
    else
	current = parent->menu->first_child;
    retval = -1;

    do {
	mc_menu = fopen("MCmenu", "wt");
	fprintf(mc_menu, "./conf/lxdialog/lxdialog --title '%s' --backtitle '" BACK_TITLE "' --menu 'Arrow keys navigate the menu.  <Enter> selects submenus --->.  Highlighted letters are hotkeys.  Pressing <Y> includes, <N> excludes, <M> modularizes features.  Press <Esc><Esc> to exit, <?> for Help.  Legend: [*] built-in  [ ] excluded  <M> module  < > module capable' %d %d %d '%s'", parent->menu->caption, rows - 4, cols - 5, rows - 14, current->name);

	print_menu(parent, mc_menu, 3);

	fprintf(mc_menu, "\n");
	fprintf(mc_menu, "exit $?\n");
	fclose(mc_menu);

	retval = system("sh MCmenu 2> MCout");
	unlink("MCmenu");

	mc_out = fopen("MCout", "rt");
	fscanf(mc_out, "%s", menu_name);
	fclose(mc_out);
	unlink("MCout");

	current = search_tree(&id_tree_root, menu_name, ID_SEARCH);

	if ((retval == 0) && current->menu->first_child) {
	    if (current->menu->flags & MENU_FLAG_EXPANDED)
		current->menu->flags -= MENU_FLAG_EXPANDED;
	    else
		current->menu->flags |= MENU_FLAG_EXPANDED;
	}

	if (retval == 512)
	    show_help(current);

	if (current && current->symbol) {
	    current0 = current;
	    switch (current->symbol->type) {
		case TYPE_BOOL:
		    if ((retval == 768) || (retval == 1280))
			current->symbol->value.char_val = 'y';
		    else if (retval == 1024)
			current->symbol->value.char_val = 'n';
		    else if (!(current->menu->first_child) && (retval == 0)) {
			if (current->symbol->value.char_val == 'y')
			    current->symbol->value.char_val = 'n';
			else
			    current->symbol->value.char_val = 'y';
		    }
		    break;
		case TYPE_TRI:
		    if (retval == 768)
			current->symbol->value.char_val = 'y';
		    else if (retval == 1280)
			current->symbol->value.char_val = 'm';
		    else if (retval == 1024)
			current->symbol->value.char_val = 'n';
		    else if (!(current->menu->first_child) && (retval == 0)) {
			if (current->symbol->value.char_val == 'm')
			    current->symbol->value.char_val = 'y';
			else if (current->symbol->value.char_val == 'y')
			    current->symbol->value.char_val = 'n';
			else
			    current->symbol->value.char_val = 'm';
		    }
		    break;
		case TYPE_INT:
		case TYPE_HEX:
		case TYPE_CHAR:
		    if (retval == 0)
			get_input(current);
		    break;
		case TYPE_STR:
		    break;
		case TYPE_OPTION:
		    if ((retval == 0) || (retval == 768) || (retval == 1280)) {
			for (option = current->menu->parent->menu->first_child;
				option; option = option->menu->next) {
			    if (option == current)
				option->symbol->value.char_val = 'y';
			    else
				option->symbol->value.char_val = 'n';
			}
		    }
		    current->menu->parent->menu->flags ^= MENU_FLAG_EXPANDED;
		    current = current->menu->parent;
		    break;
		default:
		    break;
	    }

	    for (id = current0->symbol->affects; id; id = id->next)
		set_derived_value(id->id);
	}

#ifdef DEBUG
	if (current == NULL)
	    printf("current_item NULL\n");
#endif

    } while ((retval != 256) && (retval != 65280));

    return retval;
}

int do_menuconfig()
{
    int i, retval;
    FILE *prompt_file;

    set_geometry();

    for (i = 0; i < MAX_INDENT; i++)
	indent_string[i] = ' ';

    do_menu(root_menu);

    prompt_file = fopen("MCprompt", "wt");
    fprintf(prompt_file, "./conf/lxdialog/lxdialog --backtitle '" BACK_TITLE "' --yesno 'Do you wish to save your new kernel configuration' 5 %d\n", cols - 5);
    fprintf(prompt_file, "exit $?\n");
    fclose(prompt_file);

    retval = system("sh MCprompt");
    unlink("MCprompt");

    if (retval == 0) {
	printf("\n");
	printf("\n");
	printf("Saving your kernel configuration...");
	write_config(".config", "include/linux/autoconf.h");
	printf("done.\n");
	printf("\n");
	printf("*** End of Linux kernel configuration.\n");
	printf("*** Check the top-level Makefile for additional configuration.\n");
	printf("*** Next, you must run 'make dep'.\n");
	printf("\n");
    }
    else {
	printf("\n");
	printf("\n");
	printf("Your kernel configuration changes were NOT saved.\n");
	printf("\n");
    }

    return 0;
}

