/*
 * KernelConf - A Linux Kernel Configuration System
 *
 * Distributed under GNU General Public Lisence
 *
 * Copyright (C) 2002 by Anuradha Ratnaweera
 */

#include <ctype.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/linux/version.h"

#include "kernelconf.h"
#include "deps.h"
#include "expressions.h"
#include "readsymbols.h"
#include "help.h"
#include "utils.h"
#include "tree.h"

extern IdPtr id_tree_root;
extern IdPtr first_symbol;
extern IdPtr root_menu;

static char buffer[MAX_LINE_SIZE+1];
static char *current_filename;
static int line_no;

/*
 * Print an error message in the form
 *
 *     file:line: error message
 *
 * and exit with error status 1.
 */
static void config_read_error(char *error_message)
{
    fprintf(stderr, "%s:%d: %s\n", current_filename, line_no, error_message);
    exit(1);
}

/*
 * Parse the next line from the config file and returns a pointer to the value
 * string in BUFFER.
 *
 * The tag type of the line is returned in TAG.
 */
static char *parse_line(FILE *config_file, Tag *tag)
{
    int i;
    char *bufptr, c;

    /* let's read the next line from SYMBOLFILE */
    i = 0;
    while (((c = fgetc(config_file)) != EOF) && (c != '\n') &&
	    (i < MAX_LINE_SIZE)) buffer[i++] = c;
    buffer[i] = 0;

    if ((c == EOF) && (i == 0)) {
	*tag = TAG_EOF;
	return NULL;
    }

    if (i == MAX_LINE_SIZE) {
	fprintf(stderr,
		"Warning: Line with length %d read, may have been truncated\n",
		MAX_LINE_SIZE);
    }

    /* remove comments */
    if ((bufptr = strchr(buffer, '#')))
	*bufptr = 0;

    /* if first line is space, we may have a description
     * line or a blank line */
    if (isspace(buffer[0])) {
	bufptr = string_trim(buffer);
	if (*bufptr) {
	    *tag = TAG_DESC_LINE;
	    return bufptr;
	}
	else {
	    *tag = TAG_BLANK;
	    return NULL;
	}
    }
    if (buffer[0] == 0) {
	*tag = TAG_BLANK;
	return NULL;
    }
    /* now let's handle individual tags */
    if (strncmp(buffer, "include", 7) == 0) {
	*tag = TAG_INCLUDE;
	return string_trim(buffer + 7);
    }
    if (strncmp(buffer, "symbol:", 7) == 0) {
	*tag = TAG_SYMBOL;
	return string_trim(buffer + 7);
    }
    if (strncmp(buffer, "type:", 5) == 0) {
	*tag = TAG_TYPE;
	return string_trim(buffer + 5);
    }
    if (strncmp(buffer, "value:", 6) == 0) {
	*tag = TAG_VALUE;
	return string_trim(buffer + 6);
    }
    if (strncmp(buffer, "default:", 8) == 0) {
	*tag = TAG_DEFAULT;
	return string_trim(buffer + 8);
    }
    if (strncmp(buffer, "description:", 12) == 0) {
	*tag = TAG_DESC;
	return string_trim(buffer + 12);
    }
    if (strncmp(buffer, "depends:", 8) == 0) {
	*tag = TAG_DEPENDS;
	return string_trim(buffer + 8);
    }
    if (strncmp(buffer, "derived:", 8) == 0) {
	*tag = TAG_DERIVED;
	return string_trim(buffer + 8);
    }
    if (strncmp(buffer, "menu-item:", 10) == 0) {
	*tag = TAG_MENU_ITEM;
	return string_trim(buffer + 10);
    }
    if (strncmp(buffer, "menu-caption:", 13) == 0) {
	*tag = TAG_MENU_CAPTION;
	return string_trim(buffer + 13);
    }
    if (strncmp(buffer, "menu-parent:", 12) == 0) {
	*tag = TAG_MENU_PARENT;
	return string_trim(buffer + 12);
    }
    if (strncmp(buffer, "menu-depends:", 13) == 0) {
	*tag = TAG_MENU_DEPENDS;
	return string_trim(buffer + 13);
    }
    /* unknown tag, return NULL */
    *tag = TAG_NULL;
    return NULL;
}

/*
 * Adds MENU_ITEM to the end of PARENT's children list.
 * The next element of MENU_ITEM is assumed to be NULL.
 */
static void add_menu_to_parent(IdPtr id, IdPtr parent)
{
    if (parent->menu->first_child) {
	parent->menu->last_child->menu->next = id;
	parent->menu->last_child = id;
    }
    else {
	parent->menu->first_child = id;
	parent->menu->last_child = id;
    }
    id->menu->parent = parent;
}

static MenuPtr new_menu()
{
    MenuPtr menu;

    menu = (MenuPtr)malloc(sizeof(Menu));
    menu->parent = NULL;
    menu->first_child = NULL;
    menu->last_child = NULL;
    menu->next = NULL;
    menu->caption = NULL;
    menu->depends = NULL;
    menu->flags = 0;

    return menu;
}

static IdPtr new_id_with_menu(char *name, char *config_file)
{
    IdPtr id;

    id = search_tree(&id_tree_root, name, ID_INSERT);
    if (id->menu)
	return id;
    id->menu = new_menu();
    id->config_file = config_file;

    return id;
}

static SymbolPtr new_symbol(IdPtr id)
{
    static IdPtr last_symbol = NULL;
    SymbolPtr symbol;

    symbol = (SymbolPtr)malloc(sizeof(Symbol));
    symbol->type = TYPE_BOOL;
    symbol->value.bool_val = 'n';
    symbol->depends = NULL;
    symbol->derived = NULL;
    symbol->affects = NULL;
    symbol->next = NULL;
    symbol->flags = SYMBOL_FLAG_NOT_SET;

    if (last_symbol) {
	last_symbol->symbol->next = id;
	last_symbol = id;
    }
    else
	last_symbol = id;

    return symbol;
}

static IdPtr new_id_with_symbol(char *name, char *config_file)
{
    IdPtr id;

    id = search_tree(&id_tree_root, name, ID_INSERT);
    if (id->symbol) {
#ifdef DEBUGGED
	printf("Warning: duplicate symbol: %s\n", name);
#endif
	return id;
    }
    id->symbol = new_symbol(id);
    id->config_file = config_file;
    if (first_symbol == NULL)
	first_symbol = id;

    return id;
}

static void unmark_id(IdPtr id)
{
    if (id->flags & ID_FLAG_MARK) {
	if (id->symbol == NULL) {
	    id->symbol = new_symbol(id);
	    id->flags -= ID_FLAG_MARK;
	}
	else {
	    /* we shouldn't get here */
	    fprintf(stderr, "Warning: symbol %s marked, but has a symbol\n",
		    id->name);
	}
    }
    if (id->left)
	unmark_id(id->left);
    if (id->right)
	unmark_id(id->right);
}

static void process_dependencies(IdPtr id)
{
    set_affected_ids(id);

    if (id->left)
	process_dependencies(id->left);
    if (id->right)
	process_dependencies(id->right);
}

/*
 * We have read in the configuration files and symbols and menus.  Let's do some
 * postprocessing.
 *
 * Notice that this extra pass allows us to write dependencies using yet-unread
 * symbols.
 */
int post_process()
{
    IdPtr id;

    /* make root menu and children visible */

    root_menu->menu->flags |= (MENU_FLAG_VISIBLE | MENU_FLAG_EXPANDED);
    for (id = root_menu->menu->first_child; id; id = id->menu->next)
	id->menu->flags |= MENU_FLAG_VISIBLE;

    /* traverse the tree once and create any symbol that was not read in
     * but available in dependencies. */

    unmark_id(id_tree_root);

    /* now we have a set of consistent IDs, so that we can do other
     * preprocessing stuff. */

    process_dependencies(id_tree_root);

    return 0;
}

/*
 * Reads configuration symbols, menus and dependencies from a configuration
 * file.  Calls itself recursively if include files are present.
 */
int read_symbols(char *filename)
{
    FILE *config_file;
    IdPtr id = NULL;
    Tag tag;
    char *buf, *this_file, last_line_blank, in_description;
    long location;
    static int depth = 0;
    char depends_on_modules[] = "\"m\" -> MODULES";

#ifdef DEBUG
    printf("Processing file: %s\n", filename);
#endif

    if ((config_file = fopen(filename, "rt")) == NULL) {
	perror(filename);
	exit(1);
    }

    this_file = string_duplicate(filename);
    current_filename = filename;
    line_no = 0;
    last_line_blank = 1;

    if (root_menu == NULL) {
	root_menu = new_id_with_menu("root", this_file);
	root_menu->menu->caption =
	    string_duplicate("Linux Kernel v" UTS_RELEASE " Configuration");
    }

    do {
	buf = parse_line(config_file, &tag);
	line_no++;

	switch (tag) {
	    case TAG_NULL:
		fprintf(stderr, "Warning: %s:%d: unknown tag\n",
			filename, line_no);
		break;
	    case TAG_INCLUDE:
		depth++;
		read_symbols(buf);
		depth--;
		id = NULL;
		break;
	    case TAG_SYMBOL:
		if (last_line_blank)
		    id = new_id_with_symbol(buf, this_file);
		else
		    config_read_error("Symbol defined in the middle of a block");
		break;
	    case TAG_TYPE:
		if ((id == NULL) || (id->symbol == NULL))
		    config_read_error("Type definition without symbol");
		if (strncmp(buf, "bool", 4) == 0)
		    id->symbol->type = TYPE_BOOL;
		else if (strncmp(buf, "tristate", 8) == 0) {
		    id->symbol->type = TYPE_TRI;
		    id->symbol->depends = parse_expressions(depends_on_modules,
			    id->symbol->depends);
		}
		else if (strncmp(buf, "int", 3) == 0)
		    id->symbol->type = TYPE_INT;
		else if (strncmp(buf, "hex", 3) == 0)
		    id->symbol->type = TYPE_HEX;
		else if (strncmp(buf, "char", 4) == 0)
		    id->symbol->type = TYPE_CHAR;
		else if (strncmp(buf, "str", 3) == 0)
		    id->symbol->type = TYPE_STR;
		else if (strncmp(buf, "option", 6) == 0)
		    id->symbol->type = TYPE_OPTION;
		break;
	    case TAG_VALUE:
		if (id && id->symbol)
		    id->symbol->flags |= SYMBOL_FLAG_FROZEN;
		/* do not break */
	    case TAG_DEFAULT:
		if ((id == NULL) || (id->symbol == NULL))
		    config_read_error("Value defined without symbol");
		switch (id->symbol->type) {
		    case TYPE_BOOL:
		    case TYPE_OPTION:
			if ((buf[0] == 'Y') || (buf[0] == 'y'))
			    id->symbol->value.bool_val = 'y';
			else
			    id->symbol->value.bool_val = 'n';
			break;
		    case TYPE_TRI:
			if ((buf[0] == 'Y') || (buf[0] == 'y'))
			    id->symbol->value.bool_val = 'y';
			else if ((buf[0] == 'M') || (buf[0] == 'm'))
			    id->symbol->value.bool_val = 'm';
			else
			    id->symbol->value.bool_val = 'n';
			break;
		    case TYPE_INT:
			id->symbol->value.int_val = atoi(buf);
			break;
		    case TYPE_HEX:
			sscanf(buf, "%x", &(id->symbol->value.int_val));
			break;
		    case TYPE_CHAR:
			id->symbol->value.char_val = buf[0];
			break;
		    case TYPE_STR:
			id->symbol->value.str_val = string_duplicate(buf);
			break;
		    default:
			;
		}
		break;
	    case TAG_DEPENDS:
		if ((id == NULL) || (id->symbol == NULL))
		    config_read_error("Dependency without symbol");
		if (id->symbol->flags & SYMBOL_FLAG_FROZEN)
		    config_read_error("Frozen symbol cannot have dependencies");
		id->symbol->depends = parse_expressions(buf, id->symbol->depends);
		break;
	    case TAG_DERIVED:
		if ((id == NULL) || (id->symbol == NULL))
		    config_read_error("Derived value without symbol");
		if (id->symbol->flags & SYMBOL_FLAG_FROZEN)
		    config_read_error("Frozen symbol cannot be derived");
		id->symbol->derived = parse_expressions(buf, id->symbol->derived);
		break;
	    case TAG_MENU_ITEM:
		if (last_line_blank)
		    id = new_id_with_menu(buf, this_file);
		else
		    config_read_error("Menu-item defined in the middle of a block");
		break;
	    case TAG_MENU_CAPTION:
		if (id == NULL)
		    config_read_error("Menu caption defined without menu item or symbol");
		if (id->menu == NULL)
		    id->menu = new_menu();
		if (id->menu->caption)
		    free(id->menu->caption);
		id->menu->caption = string_duplicate(buf);
		break;
	    case TAG_MENU_PARENT:
		if (id == NULL)
		    config_read_error("Menu parent defined without menu item or symbol");
		if (id->menu == NULL)
		    id->menu = new_menu();
		if (id->menu->parent == NULL) {
		    add_menu_to_parent(id, search_tree(&id_tree_root, buf,
				ID_MARK_INSERT));
		}
		break;
	    case TAG_MENU_DEPENDS:
		if (id == NULL)
		    config_read_error("Menu dependency defined without symbol or menu item");
		if (id->menu == NULL)
		    id->menu = new_menu();
		id->menu->depends = parse_expressions(buf, id->menu->depends);
		break;
	    case TAG_BLANK:
	    case TAG_EOF:
		if (id && id->menu && (id->menu->parent == NULL))
		    add_menu_to_parent(id, root_menu);
		break;
	    case TAG_DESC:
		if (id == NULL)
		    config_read_error("Description without symbol or menu item");
		in_description = 1;
		location = ftell(config_file);
		    id->desc_start = location;
		break;
	    case TAG_DESC_LINE:
		if (id == NULL)
		    config_read_error("Description line without symbol or menu item");
		if (in_description) {
		    location = ftell(config_file);
		    id->desc_length = location - id->desc_start;
		}
		else
		    config_read_error("Description line without description: tag");
		break;
	    default:
		;
	}

	if (tag == TAG_BLANK)
	    last_line_blank = 1;
	else
	    last_line_blank = 0;
	if ((tag != TAG_BLANK) && (tag != TAG_DESC) && (tag != TAG_DESC_LINE))
	    in_description = 0;

    } while (tag != TAG_EOF);

    fclose(config_file);

    /* if we are in the first file, do post-processing */
    if (depth == 0)
	post_process();

    /* do not free() THIS_FILE */

    return 0;
}

